/**
	AOK Trigger Studio (See aokts.cpp for legal conditions.)
	WINDOWS VERSION.
	mapview.cpp -- Graphical representation of AOK maps
**/

#define WIN32_LEAN_AND_MEAN
#ifndef _WIN32_IE
#define _WIN32_IE 0x0500	//IE5 for PSH_NOCONTEXTHELP flag
#define OEMRESOURCE	//OCR_NORMAL
#endif
#include <windows.h>
#include <commctrl.h>
#include "mapview.h"
#include "scen.h"
#include "settings.h"
#include <math.h>	//all the trig functions
#include <stdio.h>

/* Brushes (created with window) */
HBRUSH *tBrushes;
HBRUSH *pBrushes;
HBRUSH bWhite;

/*
	The mapview window actually draws the map to a bitmap in memory
	and then blits the bitmap to the client area when needed. This
	saves processing when resizing and scrolling the window.
*/

struct Highlight
{
	short x, y;

	struct Highlight *prev, *next;
};

struct MapViewData
{
	HDC copydc;				//DC used for saving what's under a highlight, etc.
	HBITMAP mapbmp;			//bitmap storage for map
	int bmpsize;
	int offsetx, offsety;
	int scrollMaxX, scrollMaxY;
	HWND statusbar;
	Highlight *highlights;	//LL of the highlighted points
} data;

/* Handles negatives as well */
inline int myround(double n)
{
	return int((n >= 0.0) ? n + 0.5 : n - 0.5);
}

#define pi 3.1415926536

const double sin45 = sin(pi / 4);
const double nsin45 = sin(-pi / 4);
const double root2 = sqrt(2);

/* rotates points 45deg to make a diamond from a square */

void rotate(int origin, int xi, int yi, int &xo, int &yo)
{
	double ox = xi - origin;
	double oy = yi - origin;

	/*	The origin * 2 below is an optimizization. What it's really
		doing is round(((ox + oy) * sin45 + origin * root2) * root2) so
		I just distributed the second root2.
	*/
    xo = myround((ox + oy) * sin45 * root2) + origin * 2;
    yo = myround((ox - oy) * sin45 * root2) + origin * 2;
	xo *= setts.zoom;
	yo *= setts.zoom;
}

void unrotate(int origin, int xi, int yi, unsigned &xo, unsigned &yo)
{
	double xr = xi / setts.zoom - origin * 2;
	double yr = yi / setts.zoom - origin * 2;

    xo = origin - myround((xr + yr) * nsin45 / root2);
    yo = origin - myround((xr - yr) * nsin45 / root2);
}

void PaintUnits(HDC dc)
{
	Unit *uparse;
	Player *p;
	int count, rx, ry;
	int half = scen.map.x / 2;
	RECT area;

	for (int i = 0; i < 8; i++)	//skip GAIA for now
	{
		p = scen.players + i;
		uparse = p->units.first();
		count = p->units.count();

		while (count--)
		{
			rotate(half, (int)uparse->x, (int)uparse->y, rx, ry);
			area.left = rx;
			area.right = rx + 2 * setts.zoom;
			area.top = ry;
			area.bottom = ry + 2 * setts.zoom;
			FillRect(dc, &area, pBrushes[p->color]);
			uparse++;
		}
	}

	/* TODO: Paint GAIA units here
	p = scen.players + GAIA_INDEX;
	parse = p->units.first();
	count = p->units.count();
	SelectObject(dc, bWhite);

	while (count--)
	{
		parse++;
	}
	*/

	/* Highlighting */
	for (Highlight *hparse = data.highlights; hparse; hparse = hparse->next)
	{
		rotate(half, hparse->x, hparse->y, rx, ry);
		area.left = rx;
		area.right = rx + 2 * setts.zoom;
		area.top = ry;
		area.bottom = ry + 2 * setts.zoom;
		FillRect(dc, &area, bWhite);
	}
}

void PaintMap(HDC dcdest)
{
	int half;
	unsigned y, x;
	int ry, rx;	//rotated
	Map::Terrain *parse;
	RECT area;

	/* Create a bitmap */
	data.bmpsize = 2 * scen.map.x * setts.zoom + 1;
	data.mapbmp = CreateCompatibleBitmap(dcdest, data.bmpsize, data.bmpsize);
	half = scen.map.x / 2;

	/* Do the painting. */
	SelectObject(data.copydc, data.mapbmp);
	for (x = 0; x < scen.map.x; x++)
	{
		parse = scen.map.terrain[x];
		for (y = 0; y < scen.map.y; y++, parse++)
		{
			rotate(half, x, y, rx, ry);
			area.left = rx;
			area.right = rx + 2 * setts.zoom;
			area.top = ry;
			area.bottom = ry + 2 * setts.zoom;
			FillRect(data.copydc, &area, tBrushes[parse->cnst]);
		}
	}

	PaintUnits(data.copydc);
}

/* resizes the scrollbars when window is reset or resized */
void UpdateScrollbars(HWND window, WORD width, WORD height)
{
	SCROLLINFO si;
	RECT statusSize;

	/* common init */
	si.cbSize = sizeof(si);
	si.fMask = SIF_RANGE | SIF_POS | SIF_PAGE;
	si.nMin = 0;
	si.nPos = 0;

	/* HACK: statusbar size */
	GetWindowRect(data.statusbar, &statusSize);

	/* horizontal scrollbar */
	si.nMax = data.scrollMaxX = data.bmpsize - width;
	si.nPage = si.nMax / 10;
	SetScrollInfo(window, SB_HORZ, &si, TRUE);

	/* vertical scrollbar */
	si.nMax = data.scrollMaxY =
		data.bmpsize - height + (statusSize.bottom - statusSize.top + 1);	//big hack here
	si.nPage = si.nMax / 10;
	SetScrollInfo(window, SB_VERT, &si, TRUE);
}

/* HandleRefresh: redrawing the map, not just blitting */
void HandleRefresh(HWND window, BOOL erase)
{
	HDC dc;
	RECT wndSize;
	
	/* actually redraw the map */
	dc = GetDC(window);
	if (data.mapbmp)
		DeleteObject(data.mapbmp);
	PaintMap(dc);
	ReleaseDC(window, dc);

	/* if we're doing a hard reset (erasing), update scroll bars too */
	if (erase)
	{
		GetClientRect(window, &wndSize);
		UpdateScrollbars(window, wndSize.right - wndSize.left, wndSize.bottom - wndSize.top);
	}

	/* now repaint the window */
	InvalidateRect(window, NULL, erase);
}

void HandlePaint(HWND window)
{
	HDC dc;
	PAINTSTRUCT ps;
	SCROLLINFO si;

	if (GetUpdateRect(window, NULL, FALSE))
	{
		/* Init */
		dc = BeginPaint(window, &ps);
		si.cbSize = sizeof(si);
		si.fMask = SIF_POS;

		/* scrolling offset */
		GetScrollInfo(window, SB_HORZ, &si);
		data.offsetx = (data.scrollMaxX > 0) ? -si.nPos : 0;
		GetScrollInfo(window, SB_VERT, &si);
		data.offsety = (data.scrollMaxY > 0) ? -si.nPos : 0;

		SelectObject(data.copydc, data.mapbmp);
		BitBlt(dc, data.offsetx, data.offsety, data.bmpsize, data.bmpsize,
			data.copydc, 0, 0, SRCCOPY);

		/* Cleanup */
		EndPaint(window, &ps);
	}
}

void UnhighlightPoint(HWND window, Highlight *h)
{
	/* Remove the link */
	if (h == data.highlights)
		data.highlights = h->next;

	else if (h->prev)
		h->prev->next = h->next;

	if (h->next)
		h->next->prev = h->prev;

	delete h;

	HandleRefresh(window, FALSE);
}

void UnhighlightPoint(HWND window, int x, int y)
{
	Highlight *parse, *next;

	if (x >= 0 && y >= 0)
	{
		/* Find the right highlight struct */
		for (parse = data.highlights; parse; parse = parse->next)
		{
			if (parse->x == x && parse->y == y)
				break;
		}

		/* Did we find one? */
		if (!parse)
			return;

		UnhighlightPoint(window, parse);
	}
	else
	{
		for (parse = data.highlights; parse; parse = next)
		{
			next = parse->next;	//need to do this becuase Unhighlight deletes parse
			UnhighlightPoint(window, parse);
		}
	}
}

void HighlightPoint(HWND window, int x, int y)
{
	Highlight *nh;

	if (x < 0 || y < 0)
		return;

	nh = new Highlight;
	if (!nh)
		return;

	/* Prepend the new link since we don't keep track of the last one. */
	nh->prev = NULL;
	nh->next = data.highlights;
	if (data.highlights)
		data.highlights->prev = nh;
	data.highlights = nh;

	/* Set up new highlight */
	nh->x = (short)x;
	nh->y = (short)y;

	HandleRefresh(window, FALSE);
}

#define NUM_SB_SIZES 2
int sb_sizes[NUM_SB_SIZES] =
{ 150, -1 };
const char *sb_msg = "Click here to force refresh.";

HWND makestatus(HWND parent)
{
	HWND ret;
	INITCOMMONCONTROLSEX iccx;

	iccx.dwSize = sizeof(iccx);
	iccx.dwICC = ICC_WIN95_CLASSES;

	InitCommonControlsEx(&iccx);

	ret = CreateWindow(STATUSCLASSNAME, sb_msg,
		WS_CHILD | WS_VISIBLE,
		CW_USEDEFAULT, 0, CW_USEDEFAULT, 0,
		parent,
		(HMENU)100,	//status bar uses menu parameter as ID
		NULL, NULL);

	if (ret)
		SendMessage(ret, SB_SETPARTS, NUM_SB_SIZES, (LPARAM)sb_sizes);

	return ret;
}

void OnCreate(HWND window)
{
	ColorLink *parse;
	int i;
	HDC dc;

	/* init window data */
	data.mapbmp = NULL;
	data.bmpsize = 0;
	data.statusbar = NULL;
	data.highlights = NULL;

	tBrushes = new HBRUSH[esdata.getCount(ESD_terrains)];

	for (i = 0, parse = esdata.terrains; parse; parse = (ColorLink*)parse->next, i++)
		tBrushes[i] = CreateSolidBrush(parse->ref);

	pBrushes = new HBRUSH[esdata.getCount(ESD_colors)];

	for (i = 0, parse = esdata.colors; parse; parse = (ColorLink*)parse->next, i++)
		pBrushes[i] = CreateSolidBrush(parse->ref);

	bWhite = CreateSolidBrush(0xFFFFFF);
	data.statusbar = makestatus(window);

	dc = GetWindowDC(window);
	data.copydc = CreateCompatibleDC(dc);
	ReleaseDC(window, dc);
}

void OnWM_MOUSEMOVE(HWND window, int x, int y)
{
	char text[20] = "Outside boundaries";
	unsigned rx, ry;	//un-rotated

	x -= data.offsetx;
	y -= data.offsety;
	unrotate(scen.map.x / 2, x, y, rx, ry);

	if (rx >= 0 && rx < scen.map.x && ry >= 0 && ry < scen.map.y)
		sprintf(text, "%d, %d", rx, ry);

	SendMessage(data.statusbar, SB_SETTEXT, 1, (LPARAM)text);
}

void OnWM_LBUTTONUP(HWND window, int x, int y)
{
	HWND owner = GetWindow(window, GW_OWNER);
	unsigned rx, ry;	//un-rotated

	unrotate(scen.map.x / 2, x - data.offsetx, y - data.offsety, rx, ry);

	if (rx >= 0 && rx < scen.map.x && ry >= 0 && ry < scen.map.y)
		SendMessage(owner, MAP_Click, 0, MAKELPARAM(rx, ry));
}

void OnWM_HSCROLL(HWND window, WORD type, WORD pos)
{
	SCROLLINFO si;
	int newpos;

	/* get the current scroll info */
	si.cbSize = sizeof(si);
	si.fMask = SIF_POS | SIF_RANGE;
	GetScrollInfo(window, SB_HORZ, &si);

	/* set the new position and update window */
	switch (type)
	{
	case SB_LEFT:
		newpos = si.nMin;
		break;

	case SB_RIGHT:
		newpos = si.nMax;
		break;

	case SB_LINELEFT:
		newpos = si.nPos - 1;
		break;

	case SB_LINERIGHT:
		newpos = si.nPos + 1;
		break;

	case SB_THUMBPOSITION:
	case SB_THUMBTRACK:
		newpos = pos;
		break;

	default:
		newpos = si.nPos;
	}

	InvalidateRect(window, NULL, FALSE);

	si.nPos = newpos;
	si.fMask = SIF_POS;
	SetScrollInfo(window, SB_HORZ, &si, TRUE);

}

void OnWM_VSCROLL(HWND window, WORD type, WORD pos)
{
	SCROLLINFO si;
	int newpos;

	/* get the current scroll info */
	si.cbSize = sizeof(si);
	si.fMask = SIF_POS | SIF_RANGE;
	GetScrollInfo(window, SB_VERT, &si);

	/* set the new position and update window */
	switch (type)
	{
	case SB_TOP:
		newpos = si.nMin;
		break;

	case SB_BOTTOM:
		newpos = si.nMax;
		break;

	case SB_LINEUP:
		newpos = si.nPos - 1;
		break;

	case SB_LINEDOWN:
		newpos = si.nPos + 1;
		break;

	case SB_THUMBPOSITION:
	case SB_THUMBTRACK:
		newpos = pos;
		break;

	default:
		newpos = si.nPos;
	}

	InvalidateRect(window, NULL, FALSE);

	si.nPos = newpos;
	si.fMask = SIF_POS;
	SetScrollInfo(window, SB_VERT, &si, TRUE);
}

LRESULT CALLBACK MapWndProc(HWND window, UINT msg, WPARAM wParam, LPARAM lParam)
{
	LRESULT ret;
	int i;

	switch (msg)
	{
	case WM_CREATE:
		ret = 0;
		OnCreate(window);
		break;

	case WM_PAINT:
		ret = 0;
		HandlePaint(window);
		break;

	case WM_CLOSE:
		SendMessage(GetWindow(window, GW_OWNER), MAP_Close, 0, 0);
		DestroyWindow(window);
		break;

	case WM_DESTROY:
		ret = 0;
		for (i = 0; i <  esdata.getCount(ESD_terrains); i++)
			DeleteObject(tBrushes[i]);
		for (i = 0; i < esdata.getCount(ESD_colors); i++)
			DeleteObject(pBrushes[i]);
		delete [] pBrushes;
		delete [] tBrushes;
		DeleteDC(data.copydc);
		break;

	case WM_MOUSEMOVE:
		OnWM_MOUSEMOVE(window, LOWORD(lParam), HIWORD(lParam));
		break;

	case WM_LBUTTONUP:
		OnWM_LBUTTONUP(window, LOWORD(lParam), HIWORD(lParam));
		break;

	case WM_SIZE:
		SendMessage(data.statusbar, WM_SIZE, wParam, lParam);
		UpdateScrollbars(window, LOWORD(lParam), HIWORD(lParam));
		break;

	case WM_HSCROLL:
		OnWM_HSCROLL(window, LOWORD(wParam), HIWORD(wParam));
		break;

	case WM_VSCROLL:
		OnWM_VSCROLL(window, LOWORD(wParam), HIWORD(wParam));
		break;

	case WM_NOTIFY:
		{
			const NMHDR *header = (NMHDR*)lParam;
			if (header->hwndFrom == data.statusbar && header->code == NM_CLICK)
				HandleRefresh(window, TRUE);
		}
		break;

	case MAP_HighlightPoint:
		HighlightPoint(window, wParam, lParam);
		break;

	case MAP_UnhighlightPoint:
		UnhighlightPoint(window, wParam, lParam);
		break;

	case MAP_Reset:
		HandleRefresh(window, TRUE);

	default:
		ret = DefWindowProc(window, msg, wParam, lParam);
	}

	return ret;
}

void MakeMapClass(HINSTANCE instance)
{
	WNDCLASS wc;
	
	wc.style       = CS_HREDRAW | CS_VREDRAW;
//	wc.style       = 0;
	wc.lpfnWndProc = MapWndProc;
	wc.cbClsExtra  = 0;
	wc.cbWndExtra  = 0;
	wc.hInstance   = instance;
	wc.hIcon       = NULL;
	wc.hCursor =
		(HCURSOR)LoadImage(NULL, MAKEINTRESOURCE(OCR_NORMAL), IMAGE_CURSOR, 0, 0, LR_SHARED);
	wc.hbrBackground = (HBRUSH)(COLOR_BACKGROUND);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = MapClass;

	RegisterClass(&wc);
}
