typedef struct _iobuf FILE;	//makes including <stdio.h> unnecessary

#include "datatypes.h"
#include "scen_const.h"

/** Some macros because the effect/condition player
	numbers are weird. **/
#define P_StdToEC(i) \
	if (++i == 9) \
		i = 0;
#define P_ECToStd(i) \
	if (--i == -1) \
		i = GAIA_INDEX;

/*	AOK points are y,x apparently (or at least the way
	I'm seeing them). */
struct AOKPT
{
	long y, x;
};
struct AOKRECT
{
	long top, right, bottom, left;
};

/* Triggers */

#define NUM_EFFECTS	37	//+1 for undefined
#define NUM_CONDS	24	//+1 for undefined

#pragma pack(push, 4)	//everything in effects/conditions is a long

enum TType
{
	NONE,	//just so we can have an invalid value for error checking
	TRIGGER,
	EFFECT = 0x17,
	CONDITION = 0x10
};

class ECBase	//this class acts as a type-checking base class for conditions and effects
{
protected:
	ECBase(enum TType c, long t = 0);	//called by Effect() and Condition()

public:
	long type;	//identifies type of condition/effect
	TType ttype;	//identifies EFFECT or CONDITION

	virtual int size() const = 0;

	/*	Note: The clipboard data has type and ttype reversed
		from the order AOK uses. */
	virtual void tobuffer(Buffer &b) const = 0;
};

#define MAX_UNITSEL 22

enum EffectType
{
	EFFECT_None,
	EFFECT_ChangeDiplomacy,
	EFFECT_ResearchTechnology,
	EFFECT_SendChat,
	EFFECT_PlaySound,
	EFFECT_SendTribute,
	EFFECT_UnlockGate,
	EFFECT_LockGate,
	EFFECT_ActivateTrigger,
	EFFECT_DeactivateTrigger,
	EFFECT_AIScriptGoal,
	EFFECT_CreateObject,
	EFFECT_TaskObject,
	EFFECT_DeclareVictory,
	EFFECT_KillObject,
	EFFECT_RemoveObject,
	EFFECT_ChangeView,
	EFFECT_Unload,
	EFFECT_ChangeOwnership,
	EFFECT_Patrol,
	EFFECT_DisplayInstructions,
	EFFECT_ClearInstructions,
	EFFECT_FreezeUnit,
	EFFECT_UseAdvancedButtons,
	EFFECT_DamageObject,
	EFFECT_PlaceFoundation,
	EFFECT_ChangeObjectName,
	EFFECT_ChangeObjectHP,
	EFFECT_ChangeObjectAttack,
	EFFECT_StopUnit,

	/* SWGB-only after here */
	EFFECT_SnapView,
	EFFECT_Unknown31,
	EFFECT_EnableTech,
	EFFECT_DisableTech,
	EFFECT_EnableUnit,
	EFFECT_DisableUnit,
	EFFECT_FlashObjects
};

#if (GAME == 1)
#define MAX_EFFECT EFFECT_StopUnit

#elif (GAME == 2)
#define MAX_EFFECT EFFECT_FlashObjects

#endif

class Effect : public ECBase
{
public:
	Effect();
	Effect(char *data, int size);
	void clear();
	void set(const Effect &e);
	void frombuffer(Buffer &b);
	void tobuffer(Buffer &b) const;
	bool read(FILE *in);
	void write(FILE *out);

	int size() const;
	bool check() const;

	long ai_goal;
	long amount;
	long res_type;
	long diplomacy;
	long num_sel;
	UID uid_loc;	//the selected location unit
	const UnitLink *pUnit;	//long in file, of course
	long s_player;	// GAIA = 0, Player 1 = 1, ...
	long t_player;	// GAIA = 0, Player 1 = 1, ...
	const TechLink *pTech;	//long in file, of course
	long stringid;
	long u2;
	long disp_time;
	unsigned long trig_index;
	AOKPT location;
	AOKRECT area;
	long group;	//unit group
	long utype;
	long panel;

	SString text;
	long snd_len;
	char sound[0x20];
	UID uids[MAX_UNITSEL];		//array of selected units

	static const char* types[NUM_EFFECTS];
};

enum ConditionType
{
	CONDITION_None,
	CONDITION_BringObjectToArea,
	CONDITION_BringObjectToObject,
	CONDITION_OwnObjects,
	CONDITION_OwnFewerObjects,
	CONDITION_ObjectsInArea,
	CONDITION_DestroyObject,
	CONDITION_CaptureObject,
	CONDITION_AccumulateAttribute,
	CONDITION_ResearchTehcnology,
	CONDITION_Timer,
	CONDITION_ObjectSelected,
	CONDITION_AISignal,
	CONDITION_PlayerDefeated,
	CONDITION_ObjectHasTarget,
	CONDITION_ObjectVisible,
	CONDITION_ObjectNotVisible,
	CONDITION_ResearchingTechnology,
	CONDITION_UnitsGarrisoned,
	CONDITION_DifficultyLevel,

	/* SWGB-only */
	CONDITION_OwnFewerFoundations,
	CONDITION_SelectedObjectsInArea,
	CONDITION_PoweredObjectsInArea,
	CONDITION_UnitsQueuedPastPopCap
};

#if (GAME == 1)
#define MAX_CONDITION CONDITION_DifficultyLevel

#elif (GAME == 2)
#define MAX_CONDITION CONDITION_UnitsQueuedPastPopCap

#endif

class Condition : public ECBase
{
public:
	Condition();
	Condition(char *data, int size);

	bool read(FILE *in);
	void write(FILE *out);
	void frombuffer(Buffer &b);
	void tobuffer(Buffer &b) const;
	void clear();

	int size() const;
	bool check() const;

	long	amount;
	long	res_type;
	UID		object;
	UID		u_loc;
	const UnitLink *pUnit;	//long in file, of course
	long	player;	// GAIA = 0, Player 1 = 1, ...
	const TechLink *pTech;	//long in file, of course
	long	timer;
	long	u1;
	AOKRECT	area;
	long	group;
	long	utype;
	long	ai_signal;

//internal

	static const char* types[NUM_CONDS];
};

#pragma pack(1)

class Trigger
{
	bool read(FILE *in);
	void write(FILE *out);

public:
	Trigger();
	Trigger(const Trigger &t);
	Trigger(char *data, const int size);

	long state, loop;
	bool u1, obj;
	long obj_order;
	SString description;
	char name[44];
	SVector <Effect> effects;
	SVector <Condition> conds;

	void tobuffer(char *buffer, int size) const;
	int size() const;

	/*	get_player: Returns non-GAIA player for conditions/effect in the trigger.

		Returns:
		* 0 if all players are unassigned (GAIA)
		* -1 if there is a mix of non-GAIA players
		* EC player index if above are false
	*/
	int get_player();

	/*	change_player: Changes all non-GAIA players to specified player. */
	void change_player(unsigned to);

	friend class Scenario;
};

#pragma pack(pop)
