/**
	AOK Trigger Studio (See aokts.cpp for legal conditions.)
	WINDOWS VERSION.

	utilunit.cpp: Various utility functions for units management.
**/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "scen.h"
#include "utilunit.h"
#include "utilui.h"
#include "resource.h"
#include "settings.h"
#include <stdio.h>

/** Sorting **/

char *sorts[] = { "ID", "Name" };

int compare_id(const void *elem1, const void *elem2)
{
	int sort1, sort2;

	sort1 = (*(Unit**)elem1)->ident;
	sort2 = (*(Unit**)elem2)->ident;

	if (sort1 > sort2)
		return 1;
	else if (sort1 == sort2)
		return 0;
	else
		return -1;
}

int compare_name(const void *elem1, const void *elem2)
{
	int ret;

#ifdef FASTSORT //broken

	int sort1, sort2;

	Unit *u1 = *(Unit**)elem1;
	Unit *u2 = *(Unit**)elem2;
	sort1 = units_sorted[u1->getType()->id];
	sort2 = units_sorted[u2->getType()->id];

	if (sort1 == sort2)
		ret = compare_id(elem1, elem2); //now resort to id
	else if (sort1 > sort2)
		ret = 1;
	else
		ret = -1;

#else

	Unit *u1 = *(Unit**)elem1;
	Unit *u2 = *(Unit**)elem2;

	ret = _stricoll(u1->getType()->name, u2->getType()->name);

	if (ret == 0)
		ret = compare_id(elem1, elem2); //now resort to id

#endif

	return ret;
}

/** Utility Functions **/

/*
	MakePtrArray: Makes an array of Unit*'s for sorting from an array of Units.
*/
inline const Unit **MakePtrArray(const Unit *list, unsigned count)
{
	const Unit **ptrs, **pp;

	ptrs = new const Unit*[count];
	pp = ptrs;

	while (count--)
		*pp++ = list++;

	return ptrs;
}

/*
	UnitList_Fill: Fill a listbox from an array of struct Unit's.

	Note: Will SETITEMDATA to the index in the array.
*/
void UnitList_Fill(HWND listbox, enum Sorts sorttype,
				   const Unit *list, unsigned count)
{
	const Unit *u;
	char string[48];
	unsigned int i;
	const Unit **sorted, **s_parse;

	SendMessage(listbox, LB_RESETCONTENT, 0, 0);

	/* sort with CRT's qsort() func. */
	sorted = MakePtrArray(list, count);
	switch (sorttype)
	{
	case SORT_ID:
		qsort(sorted, count, sizeof(Unit*), compare_id);
		break;
	case SORT_Name:
		qsort(sorted, count, sizeof(Unit*), compare_name);
		break;
	}

	/* then add the sorted items to the listbox. */
	s_parse = sorted;
	for (i = 0; i < count; i++)
	{
		u = *s_parse++;

		_snprintf(string, sizeof(string), "%d: %s",
			u->ident, u->getType()->name);

		SendMessage(listbox, LB_ADDSTRING, 0, (LPARAM)string);

		/* Now set the item's data to its index in the array. */
		SendMessage(listbox, LB_SETITEMDATA, i, u - list);
	}

	delete [] sorted;
}

/*
	UnitList_InsertItem: Inserts a new item in the listbox (in sort order)
		and returns its position.
*/
int UnitList_InsertItem(HWND listbox, enum Sorts sorttype,
						const Unit *list, unsigned count, const Unit *which)
{
	char string[48];
	int ret = -1;
	const Unit **sorted, **s_parse;

	/* sort with CRT's qsort() func. */
	sorted = MakePtrArray(list, count);
	switch (sorttype)
	{
	case SORT_ID:
		qsort(sorted, count, sizeof(Unit*), compare_id);
		break;
	case SORT_Name:
		qsort(sorted, count, sizeof(Unit*), compare_name);
		break;
	}

	_snprintf(string, sizeof(string), "%d: %s",
		which->ident, which->getType()->name);

	s_parse = sorted;
	while (count--)
	{
		if (*s_parse == which)
		{
			ret = s_parse - sorted;
			SendMessage(listbox, LB_INSERTSTRING, ret, (LPARAM)string);
			SendMessage(listbox, LB_SETITEMDATA, ret, which - list);
		}

		s_parse++;
	}

	delete [] sorted;

	return ret;
}

void UnitList_FillGroup(HWND typebox, UnitGroupLink *group)
{
	unsigned count;
	const UnitLink **c_parse;
	int index;

	SendMessage(typebox, LB_RESETCONTENT, 0, 0);

	if (group)
	{
		count = group->count;
		c_parse = group->list;

		if (!c_parse)
			return;

		while (count--)
		{
			index = SendMessage(typebox, LB_ADDSTRING, 0, (LPARAM)(*c_parse)->name);
			SendMessage(typebox, LB_SETITEMDATA, index, (LPARAM)*c_parse++);
		}
	}
	else
	{
		for (const UnitLink *l_parse = esdata.units;
			l_parse; l_parse = (UnitLink*)l_parse->next)
		{
			index = SendMessage(typebox, LB_ADDSTRING, 0, (LPARAM)l_parse->name);
			SendMessage(typebox, LB_SETITEMDATA, index, (LPARAM)l_parse);
		}
	}
}

/*
	UnitList_FillSelected: Fill a listbox with an array of pointers to Unit structs.

	Note: Will SETITEMDATA to the index of the unit in the array.
*/
void UnitList_FillSelected(HWND listbox, const SVector <Unit> *vector,
						   unsigned int *selected, unsigned int count, enum Sorts sorttype)
{
	const Unit *u_parse;
	char string[48];
	unsigned int i, index, v_count;
	unsigned int *s_parse;
	int c_count;
	const Unit **sorted, **sort_parse;
	bool multi;	//multiple-selection listbox?

	u_parse = vector->first();
	v_count = vector->count();
	multi = (GetWindowLongPtr(listbox, GWL_STYLE) & LBS_MULTIPLESEL) != 0;

	/* Sort time! */
	sorted = MakePtrArray(u_parse, v_count);
	switch (sorttype)
	{
	case SORT_ID:
		qsort(sorted, v_count, sizeof(Unit*), compare_id);
		break;
	case SORT_Name:
		qsort(sorted, v_count, sizeof(Unit*), compare_name);
		break;
	}

	SendMessage(listbox, LB_RESETCONTENT, 0, 0);
	sort_parse = sorted;

	for (i = 0; i < v_count; i++)
	{
		const Unit *u = *sort_parse++;

		sprintf(string, "%d: %s", u->ident, u->getType()->name);
		index = SendMessage(listbox, LB_ADDSTRING, 0, (LPARAM)string);

		/* Set the item's data to its index in the array. */
		SendMessage(listbox, LB_SETITEMDATA, index, u - vector->first());

		/* if the unit is one of the selected, select it */
		if (multi)
		{
			s_parse = selected;
			c_count = count;

			while (c_count--)
			{
				if (u->ident == *s_parse)
					SendMessage(listbox, LB_SETSEL, TRUE, index);
				s_parse++;
			}
		}
		else if (selected && u->ident == *selected)
			SendMessage(listbox, LB_SETCURSEL, index, 0);
	}
}

/* The Unit Selector */

void Init(HWND dialog, LPARAM lParam)
{
	Unit *u_parse;
	int count, i;
	struct UnitEdit *ue;

	SetWindowLongPtr(dialog, DWLP_USER, lParam);
	ue = (UnitEdit*)lParam;

	/* Initialize controls. */
	HWND control = GetDlgItem(dialog, IDC_US_SORT);
	SendMessage(control, CB_ADDSTRING, 0, (LPARAM)sorts[0]);
//	SendMessage(control, CB_ADDSTRING, 0, (LPARAM)sorts[1]);
	SendMessage(control, CB_SETCURSEL, 0, 0);
	FILL_CB(IDC_US_PLAYER, Player::names, NUM_PLAYERS);

	/* Temporarily convert the player id to standard value. (See trigger.h.) */
	if (ue->player >= 0)
	{
		P_ECToStd(ue->player);
	}

	if (ue->count > 0)
	{
		/* Find the owner player. */
		for (i = 0; i < NUM_PLAYERS && ue->player == -1; i++)
		{
			u_parse = scen.players[i].units.first();

			for (count = scen.players[i].units.count(); count; count--, u_parse++)
			{
				if (u_parse->ident == *ue->ids)
				{
					ue->player = i;
					break;
				}
			}
		}

		SendMessage(control, CB_SETCURSEL, ue->player, 0);	//player selection

		UnitList_FillSelected(GetDlgItem(dialog, IDC_US_UNITS),
			&scen.players[ue->player].units,
			(unsigned int *)ue->ids, ue->count, SORT_ID);
	}
}

char warningSelchange[] =
"This operation will clear the current selection. Continue?";

void HandleCommand(HWND dialog, WORD code, WORD id, HWND control)
{
	/* Used by most commands */
	struct UnitEdit *ue = (UnitEdit*)GetWindowLongPtr(dialog, DWLP_USER);

	if (id == IDOK && code == BN_CLICKED)
	{
		HWND listbox = GetDlgItem(dialog, IDC_US_UNITS);
		int buffer[MAX_UNITSEL];
		int index;
		bool multi;	//is this IDD_UNITSEL2 or just IDD_UNITSEL1?
		multi = (GetWindowLongPtr(listbox, GWL_STYLE) & LBS_MULTIPLESEL) != 0;

		if (multi)
		{
			ue->count = (short)SendMessage(listbox, LB_GETSELCOUNT, 0, 0);
			SendMessage(listbox, LB_GETSELITEMS, MAX_UNITSEL, (LPARAM)&buffer);
		}
		else
		{
			*buffer = SendMessage(listbox, LB_GETCURSEL, 0, 0);
			ue->count = (*buffer != LB_ERR);	//boolean magic!
		}

		/* Update the uids. We don't have to worry about clearing the
		   unused ones because only the used ones are written to the scen. */
		for (int i = 0; i < ue->count; i++)
		{
			index = SendMessage(listbox, LB_GETITEMDATA, buffer[i], 0);
			ue->ids[i] = scen.players[ue->player].units.at(index)->ident;
		}

		ue->player = (char)SendDlgItemMessage(dialog, IDC_US_PLAYER, CB_GETCURSEL, 0, 0);

		P_StdToEC(ue->player);

		EndDialog(dialog, 1);
	}
	else if (id == IDCANCEL && code == BN_CLICKED)
	{
		P_StdToEC(ue->player);
		EndDialog(dialog, 0);
	}
	else if (id == IDC_US_PLAYER && code == CBN_SELCHANGE)
	{
		int player, code = IDYES;
		HWND unitbox;

		/* Init */
		player = SendMessage(control, CB_GETCURSEL, 0, 0);
		unitbox = GetDlgItem(dialog, IDC_US_UNITS);

		/* Warn user if there are currently items selected */
		if (SendMessage(unitbox, LB_GETSELCOUNT, 0, 0) > 0)
			code = MessageBox(dialog, warningSelchange,
								"Player Change", MB_ICONWARNING | MB_YESNO);

		/* Take appropriate action */
		if (code == IDYES)
		{
			ue->player = player;
			UnitList_FillSelected(unitbox,
				&scen.players[player].units, NULL, 0, SORT_ID);
		}
		else
			SendMessage(control, CB_SETCURSEL, ue->player, 0);
	}
}

INT_PTR CALLBACK SelDlgProc(HWND dialog, UINT msg, WPARAM wParam, LPARAM lParam)
{
	INT_PTR ret = FALSE;

	switch (msg)
	{
	case WM_INITDIALOG:
		Init(dialog, lParam);
		ret = TRUE;
		break;

	case WM_COMMAND:
		HandleCommand(dialog, HIWORD(wParam), LOWORD(wParam), (HWND)lParam);
		break;
	}

	return ret;
}
